//---------------------------------------------------------------------------

#pragma hdrstop

#define _EXPORTING
#include "..\miniscope\device\MiniscopeDevice.h"
#include "..\miniscope\device\MiniscopeDeviceCapabilities.h"
#include <assert.h>
#include "ArmScope.h"
#include <algorithm>	// needed by Utils::in_group
#include "Utils.h"
#include <string>
#include <fstream> 
#include <json/json.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)

static const struct S_DEVICE_DLL_INTERFACE dll_interface =
	{DLL_INTERFACE_MAJOR_VERSION, DLL_INTERFACE_MINOR_VERSION};

// callback ptrs
CALLBACK_LOG lpLogFn = NULL;
CALLBACK_CONNECT lpConnectFn = NULL;
CALLBACK_DATA lpDataFn = NULL;
CALLBACK_TRIGGER lpTriggerFn = NULL;


void *callbackCookie;	///< used by upper class to distinguish library instances when receiving callbacks

void __stdcall GetMiniscopeInterfaceDescription(struct S_DEVICE_DLL_INTERFACE* interf)
{
	interf->majorVersion = dll_interface.majorVersion;
	interf->minorVersion = dll_interface.minorVersion;
}

void Log(char* txt)
{
	if (lpLogFn)
		lpLogFn(callbackCookie, txt);
}

void Data(short* data, int size)
{
	if (lpDataFn)
		lpDataFn(callbackCookie, data, size);
}

void Connect(int state, char *szMsgText)
{
	if (lpConnectFn)
		lpConnectFn(callbackCookie, state, szMsgText);
}

void Trigger(void)
{
	if (lpTriggerFn)
		lpTriggerFn(callbackCookie);
}

void __stdcall SetCallbacks(void *cookie, CALLBACK_LOG lpLog, CALLBACK_CONNECT lpConnect,
	CALLBACK_DATA lpData, CALLBACK_TRIGGER lpTrigger)
{
	assert(cookie && lpLog && lpConnect && lpData && lpTrigger);
	lpLogFn = lpLog;
	lpConnectFn = lpConnect;
	lpDataFn = lpData;
	lpTriggerFn = lpTrigger;
	callbackCookie = cookie;
	lpLogFn(callbackCookie, "armscope.dll loaded. Yay!");

	armScope.callbackLog = Log;
	armScope.callbackData = Data;
	armScope.callbackConnect = Connect;
	armScope.callbackTrigger = Trigger;
}

static enum E_TRIGGER_TYPE eTrigger[] = { TRIGGER_MANUAL, TRIGGER_SLOPE_UP, TRIGGER_SLOPE_DOWN, TRIGGER_CONTINUOUS };
static enum E_COUPLING_TYPE eCoupling[] = { COUPLING_DC, COUPLING_AC };
static enum E_ACQUISITION_MODE eAcqMode[] = { ACQUISITION_1SHOT, ACQUISITION_CONTINUOUS };
static unsigned int iBuf[] = { 512, 1024, 2048, 4096, 8192, 14336 };
static float fSens[] = { 0.000547, 0.001602, 0.004766, 0.014335, 0.042969, 0.128906 };
static float fSampl[] = { 2.048e-3, 1.024e-3, 5.12e-4, 2.56e-4, 1.28e-4, 6.4e-5,
	3.2e-5, 1.6e-5, 7.971087e-6, 3.964677e-6, 1.96147e-6 };

/** \brief Calibration data for single sensitivity range
*/
static struct S_CALIBRATION
{
	int offset;
	float gain;
} calibration[sizeof(fSens)/sizeof(fSens[0])];

void __stdcall GetDeviceCapabilities(struct S_SCOPE_CAPABILITIES **caps)
{
	static struct S_SCOPE_CAPABILITIES capabilities =
	{
		8,		//bits/sample
		sizeof(fSens)/sizeof(fSens[0]), fSens,
		sizeof(fSampl)/sizeof(fSampl[0]), fSampl,
		sizeof(eTrigger)/sizeof(eTrigger[0]), eTrigger,
		sizeof(eAcqMode)/sizeof(eAcqMode[0]), eAcqMode,
		sizeof(eCoupling)/sizeof(eCoupling[0]), eCoupling,
		sizeof(iBuf)/sizeof(iBuf[0]), iBuf,
		0		//unsigned int iOffset;               ///< is offset setting available?
	};
	*caps = &capabilities;
}

void __stdcall ShowSettings(HANDLE parent)
{
	MessageBox(parent, "Armscope.dll has no additional settings.", "Device DLL", MB_ICONINFORMATION);
}

int __stdcall SetSensitivity(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(fSens)/sizeof(fSens[0])))
		return -1;
	return armScope.SetSensitivity(iId);
}

int __stdcall SetSamplingPeriod(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(fSampl)/sizeof(fSampl[0])))
		return -1;
	return armScope.SetSampling(sizeof(fSampl)/sizeof(fSampl[0]) - iId - 1);
}

int __stdcall SetTriggerSource(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(eTrigger)/sizeof(enum E_TRIGGER_TYPE)))
		return -1;
	return armScope.SetTriggerType(eTrigger[iId]);
}

int __stdcall SetTriggerLevel(int level)
{
	return armScope.SetTriggerLevel(level);	
}

int __stdcall SetAcqMode(int iId)
{
	return 0;
}

int __stdcall SetCouplingType(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(eCoupling)/sizeof(eCoupling[0])))
		return -1;
	return armScope.SetCouplingType(eCoupling[iId]);
}

int __stdcall SetBufferSize(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(iBuf)/sizeof(unsigned int)))
		return -1;
	return armScope.SetBufferSize(iBuf[iId]);
}

int __stdcall SetOffset(float fOffset)
{
	return 0;
}

int __stdcall Connect(void)
{
	return armScope.Connect();
}

int __stdcall Disconnect(void)
{
	return armScope.Disconnect();
}

int __stdcall AcquireStart(void)
{
	armScope.SetTriggerType(TRIGGER_CONTINUOUS);
	return armScope.ManualTrigger();
}

int __stdcall AcquireStop(void)
{
	return armScope.SetTriggerType(TRIGGER_MANUAL);
}

int __stdcall ManualTrigger(void)
{
	return armScope.ManualTrigger();
}

static bool bSettingsReaded = false;

static int GetDefaultSettings(struct S_SCOPE_SETTINGS* settings)
{
	settings->iSensitivity = 0;
	settings->iSamplingPeriod = 0;
	settings->eTriggerType = TRIGGER_MANUAL;
	settings->iTriggerLevel = 0;
	settings->eAcqModeType = ACQUISITION_1SHOT;
	settings->eCouplingType = COUPLING_DC;
	settings->iBufferSize = 512;
	settings->iOffset = 0;
	for (unsigned int i=0; i<sizeof(calibration)/sizeof(calibration[0]); i++)
	{
		calibration[i].offset = 0;
		calibration[i].gain = 1.0;
	}
	bSettingsReaded = true;
	return 0;
}

int __stdcall GetMiniscopeSettings(struct S_SCOPE_SETTINGS* settings)
{
	std::string path = Utils::GetDllPath();
	path = Utils::ReplaceFileExtension(path, ".cfg");
	if (path == "")
		return GetDefaultSettings(settings);

	Json::Value root;   // will contains the root value after parsing.
	Json::Reader reader;

	std::ifstream ifs(path.c_str());
	std::string strConfig((std::istreambuf_iterator<char>(ifs)), std::istreambuf_iterator<char>());
	ifs.close();

	bool parsingSuccessful = reader.parse( strConfig, root );
	if ( !parsingSuccessful )
		return GetDefaultSettings(settings);

	settings->iSensitivity = root.get("Sensitivity", 0).asInt();
	if (settings->iSensitivity >= sizeof(fSens)/sizeof(fSens[0]))
		settings->iSensitivity = 0;

	settings->iSamplingPeriod = root.get("SamplingPeriod", 0).asInt();
	if (settings->iSamplingPeriod >= sizeof(fSampl)/sizeof(fSampl[0]))
		settings->iSamplingPeriod = 0;

	int trigger = root.get("TriggerType", TRIGGER_MANUAL).asInt();
	if (Utils::in_group(trigger, eTrigger))
		settings->eTriggerType = (E_TRIGGER_TYPE)trigger;
	else
		settings->eTriggerType = TRIGGER_MANUAL;

	settings->iTriggerLevel = root.get("TriggerLevel", 0).asInt();
	if (settings->iTriggerLevel < -100 || settings->iTriggerLevel > 100)
		settings->iTriggerLevel = 0;

	int acqmode = root.get("AcqModeType", ACQUISITION_1SHOT).asInt();
	if (Utils::in_group(acqmode, eAcqMode))
		settings->eAcqModeType = (E_ACQUISITION_MODE)acqmode;
	else
		settings->eAcqModeType = ACQUISITION_1SHOT;

	int coupling = root.get("CouplingType", COUPLING_DC).asInt();
	if (Utils::in_group(coupling, eCoupling))
		settings->eCouplingType = (E_COUPLING_TYPE)coupling;
	else
		settings->eCouplingType = COUPLING_DC;

	settings->iBufferSize = root.get("BufferSize", 512).asInt();
	if (!Utils::in_group(settings->iBufferSize, iBuf))
		settings->iBufferSize = iBuf[0];

	settings->iOffset = root.get("Offset", 0).asInt();

	// read calibration data also
	const Json::Value rootcal = root["Calibration"];	
	for (unsigned int i=0; i<sizeof(calibration)/sizeof(calibration[0]); i++)
	{
		calibration[i].offset = rootcal[i].get("Offset", 0).asInt();
		calibration[i].gain = rootcal[i].get("Gain", 1.0).asDouble();
	}
	bSettingsReaded = true;
	return 0;
}

int __stdcall SaveMiniscopeSettings(struct S_SCOPE_SETTINGS* settings)
{
    Json::Value root;
	Json::StyledWriter writer;
	root["Sensitivity"] = settings->iSensitivity;
	root["SamplingPeriod"] = settings->iSamplingPeriod;
	root["TriggerType"] = settings->eTriggerType;
	root["TriggerLevel"] = settings->iTriggerLevel;
	root["AcqModeType"] = settings->eAcqModeType;
	root["CouplingType"] = settings->eCouplingType;
	root["BufferSize"] = settings->iBufferSize;
	root["Offset"] = settings->iOffset;

	// add calibration data also
	const Json::Value rootcal = root["Calibration"];
	for (unsigned int i=0; i<sizeof(calibration)/sizeof(calibration[0]); i++)
	{
		  root["Calibration"][i]["Offset"] = calibration[i].offset;
		  root["Calibration"][i]["Gain"] = calibration[i].gain;
	}

	std::string outputConfig = writer.write( root );

	std::string path = Utils::GetDllPath();
	path = Utils::ReplaceFileExtension(path, ".cfg");
	if (path == "")
		return -1;

    std::ofstream ofs(path.c_str());
    ofs << outputConfig;
	ofs.close();

	return 0;
}

int __stdcall StoreCalibration(int range, int offset, float gain)
{
	if (range < 0 || range >= (int)(sizeof(fSens)/sizeof(fSens[0])))
		return -1;
	calibration[range].offset = offset;
	calibration[range].gain = gain;
	return 0;
}

int __stdcall GetCalibration(int range, int* const offset, float* const gain)
{
	if (range < 0 || range >= (int)(sizeof(fSens)/sizeof(fSens[0])))
		return -1;
	struct S_SCOPE_SETTINGS settings;
	if (bSettingsReaded == false)
	{
		int status = GetMiniscopeSettings(&settings);
		if (status)
			return status;
	}
	*offset = calibration[range].offset;
	*gain = calibration[range].gain;
	return 0;
}

